# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""spike detection and extraction

Copyright (C) 2013 Dan Meliza <dmeliza@gmail.com>
Created Fri Jul 12 14:05:16 2013
"""
from typing import Iterable, Union, Tuple, Optional, Iterator, Callable
import numpy as np

numeric = Union[int, float, np.number]


def filter_times(
    times: Iterable[numeric], min: numeric, max: numeric
) -> Tuple[numeric]:
    return tuple(t for t in times if (t > min) and (t < max))


def peak_idx(spikes: np.ndarray):
    """Return index of maximum negative value"""
    return spikes.argmax(-1)


def trough_idx(spikes: np.ndarray):
    """Return index of maximum positive value"""
    return spikes.argmin(-1)


def upswing_idx(spikes: np.ndarray):
    """Return index of the point midway between the trough and the following peak"""
    _trough = trough_idx(spikes)
    # need to find the local maximum after the trough, not the global max
    raise NotImplementedError


def realign_spikes(
    times: Iterable[numeric],
    spikes: np.ndarray,
    *,
    upsample: int = 1,
    align_by: Callable[[np.ndarray], Union[np.ndarray, int]] = peak_idx,
    jitter: int = 3,
    expected_peak: Optional[int] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Realign spikes to their peaks (or some other feature) using bandwidth-limited resampling

    times    : one-dimensional array of spike times, in units of samples
    spikes   : array of spike waveforms, with dimensions (nspikes, npoints)
    upsample : integer, the factor by which to upsample the data for realignment (must be greater than 1)
    align_by : function that returns the alignment point in each spike (see below)
    jitter   : samples (in original data) to search for the alignment point
    expected_peak : if supplied, searches around this index (in original data) for the peak. If not supplied,
                    uses align_by on the mean spike

    The align_by function takes a 1D or 2D array and returns the index
    corresponding to the alignment point in the spike. If the array is 1D, the
    function should return a single number. If the array is 2D, the function
    must return an array. For example, peak_idx is lambda x: x.argmax(-1)

    Returns (times, spikes), with the sampling rate increased by a factor of
    upsample.

    """
    if not (isinstance(upsample, int) and upsample > 0):
        raise ValueError("Upsampling factor must be a positive integer")
    nevents, nsamples = spikes.shape

    if expected_peak is None:
        # infer expected peak from the mean spike
        expected_peak = align_by(np.mean(spikes, 0))
    expected_peak *= upsample

    if upsample > 1:
        spikes = fftresample(spikes, int(nsamples * upsample), reflect=True)
    # find peaks within upsample samples of mean peak
    window = upsample * jitter
    # output array is clipped by jitter on both ends and then upsampled
    nshifted = (nsamples - 2 * jitter) * upsample
    shifted = np.zeros((nevents, nshifted))
    # offset is the position of the peak (or trough) within the search window
    search = slice(expected_peak - window, expected_peak + window + 1)
    start = align_by(spikes[:, search])
    for i, spike in enumerate(spikes):
        shifted[i, :] = spike[start[i] : start[i] + nshifted]
    # spike times are shifted by offset - search window
    return (np.asarray(times) * upsample + start - window, shifted)


def fftresample(S: np.ndarray, npoints: int, *, reflect: bool = False) -> np.ndarray:
    """Resample an array of waveforms using discrete fourier transform. The
    signal is transformed in the fourier domain and then padded or truncated to
    the desired size. This should be equivalent to a sinc resampling.

    S: a 2D array of waveforms (nspikes, nsamples)
    npoints: the desired duration of the resampled signal
    reflect: if True, pad the sample with reflected copies on either end to suppress edge effects

    """
    from numpy.fft import rfft, irfft

    if reflect:
        Srev = S[:, ::-1]
        S = np.column_stack([Srev, S, Srev])
        npoints = npoints * 3
    Sf = rfft(S, axis=1)
    Srs = (1.0 * npoints / S.shape[1]) * irfft(Sf, npoints, axis=1)
    if reflect:
        npoints = npoints // 3
        return Srs[:, npoints : npoints * 2]
    else:
        return Srs


def trim_waveforms(
    spikes: Iterable[np.ndarray], times: Iterable[int], *, peak_t: int, n_rise: int
) -> Iterator[Tuple[int, np.ndarray]]:
    """Trims spike waveforms to remove overlapping peaks.

    The peaks() function is used to extract waveforms in a window surrounding
    peak times. The windows for successive peaks may overlap each other if the
    peak times are closer together than the window size, which can result in
    secondary peaks occuring later in the window corresponding to the subsequent
    peak(s). If this is not desired, this function can be used to trim the
    samples after n_rise before the next peak.

    times: the times of the spikes (in samples)
    spikes: a list of spike waveforms, or a 2D array with dimensions (nspikes, nsamples)
    peak_t: the index of the peak in the spike waveforms
    n_rise: trim up to this many samples before the subsequent peak

    Note that there is nothing to prevent a spike from being completely trimmed
    if the subsequent spike is very close and n_rise is too large. In practice
    this is highly unlikely.

    """
    diffs = np.diff(times)
    for t, dt, s in zip(times, diffs, spikes):
        i_last = min(
            peak_t + dt - n_rise,
            # s[peak_t:].argmin() + peak_t,
            s.size,
        )
        yield t, s[:i_last]
    # the last spike
    s = spikes[-1]
    # i_last = min(s[peak_t:].argmin() + peak_t, s.size)
    yield times[-1], s[: s.size]


def runlength_encode(arr: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Run length encoding of arr. Returns (runlengths, startpositions, values)"""
    # from stackoverflow 1066758
    n = len(arr)
    if n == 0:
        return (None, None, None)
    else:
        y = np.array(arr[1:] != arr[:-1])  # pairwise unequal (string safe)
        i = np.append(np.where(y), n - 1)  # must include last element posi
        z = np.diff(np.append(-1, i))  # run lengths
        p = np.cumsum(np.append(0, z))[:-1]  # positions
        return (z, p, arr[i])



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""specialized functions for intracellular data"""
from typing import Tuple, Optional, Iterator
from collections import namedtuple
import logging
import numpy as np

log = logging.getLogger("quickspikes")

Spike = namedtuple(
    "Spike",
    [
        "peak_t",
        "peak_V",
        "takeoff_t",
        "takeoff_V",
        "trough_t",
        "trough_V",
        "half_rise_t",
        "half_decay_t",
        "max_dV_t",
        "max_dV",
        "min_dV_t",
        "min_dV",
    ],
)


def spike_shape(
    spike: np.ndarray,
    dt: float,
    *,
    deriv_thresh: float = 10.0,
    t_baseline: float = 2.0,
    min_rise: float = 0.25,
) -> Optional[Spike]:
    """Computes spike shape features:

    takeoff: the voltage/time when the derivative of the waveform exceeds
            `deriv_thresh` standard deviations over a baseline period (
            `t_baseline`) for at least `min_rise` ms.

    trough: The minimum value between the spike peak and the first sample where
            the derivative becomes greater than zero for at least `min_rise`
            ms. If the derivative never crosses this threshold, then it
            this is the global minimum after the peak.

    half_rise: the half-way point between the takeoff and the peak on the rising phase
    half_decay: the halfway point between the peak and the takeoff on the decay phase
    max_deriv: the maximum dV/dt on the rising phase
    min_deriv: the minimum dV/dt on the falling phase

    The default values work well for an intracellular spike recorded at 50 kHz
    with a clearly defined onset and a width 1-2 ms. If the spikes are narrower,
    min_rise should be reduced. If the spikes are too close together, it may be
    difficult to establish a good baseline for calculating the threshold.

    All features are returned as both a time and a value. If no takeoff point
    can be found, then half_rise and half_decay are undefined. If the spike is
    very badly formed (e.g. the peak is at the very beginning or end), then None
    is returned.

    """
    from quickspikes.spikes import find_run
    from quickspikes.tools import runlength_encode

    n_baseline = int(t_baseline / dt)
    min_rise = int(min_rise / dt)
    peak_ind = spike.argmax()
    if peak_ind in (0, spike.size - 1):
        return None
    peak_v = spike[peak_ind]
    deriv = np.gradient(spike, dt)
    # trough
    dV0_ind = find_run(deriv[peak_ind:], thresh=0, min_run=min_rise) or spike.size
    trough_ind = spike[peak_ind : peak_ind + dV0_ind].argmin()
    # min and max dV/dt
    dVmax_ind = deriv[:peak_ind].argmax()
    dVmin_ind = deriv[peak_ind:].argmin()

    # takeoff
    mdV = deriv[:n_baseline].mean()
    sdV = deriv[:n_baseline].std()
    thresh = mdV + deriv_thresh * sdV
    n, p, v = runlength_encode(deriv[:peak_ind] > thresh)
    pp = p[(n >= min_rise) & v]
    try:
        takeoff_ind = pp[-1]
    except IndexError:
        takeoff_t = takeoff_v = half_rise_t = half_decay_t = None
    else:
        takeoff_v = spike[takeoff_ind]
        takeoff_t = (peak_ind - takeoff_ind) * dt
        half_ampl = (peak_v + takeoff_v) / 2
        half_rise_ind = find_run(spike[:peak_ind], thresh=half_ampl, min_run=1) or 0
        half_rise_t = (peak_ind - half_rise_ind) * dt
        half_decay_ind = (
            find_run(-spike[peak_ind:], thresh=-half_ampl, min_run=2)
            or spike.size - peak_ind
        )
        half_decay_t = half_decay_ind * dt
    return Spike(
        peak_t=peak_ind * dt,
        peak_V=peak_v,
        takeoff_t=takeoff_t,
        takeoff_V=takeoff_v,
        trough_t=trough_ind * dt,
        trough_V=spike[peak_ind + trough_ind],
        half_rise_t=half_rise_t,
        half_decay_t=half_decay_t,
        max_dV=deriv[dVmax_ind],
        max_dV_t=(peak_ind - dVmax_ind) * dt,
        min_dV=deriv[dVmin_ind + peak_ind],
        min_dV_t=dVmin_ind * dt,
    )


class SpikeFinder:
    """Dynamic spike finder for intracellular voltage data.

    Intracellular recordings present some special challenges for detecting
    spikes. The data are not centered, and spike waveforms can change over the
    course of stimulation. The approach here is to use the first spike in the
    recording, which is typically the largest and has the best-defined onset, to
    set the threshhold for the window discriminator.

    `n_rise`: the approximate rise time of the spikes to be detected, in samples
    `n_before`: the number of samples before the spike peaks to analyze/extract
    `n_after`: the number of samples after the spike peaks to analyze/extract
    """

    def __init__(self, n_rise: int, n_before: int, n_after: int):
        self.n_rise = n_rise
        self.n_before = n_before
        self.n_after = n_after

    def calculate_threshold(
        self,
        V: np.ndarray,
        *,
        thresh_rel: float = 0.35,
        thresh_min: float = -50,
        deriv_thresh: float = 10.0,
    ) -> Optional[Spike]:
        """Calculate the detection threshold from the amplitude of the first spike in V.

        If no spike can be detected in the signal, returns None. Otherwise, the
        instance's threshold is set at `thresh_rel` times the amplitude of the
        first spike, or `thresh_min` (whichever is greater), and the spike shape
        is returned.

        """
        self.spike_thresh = self.first_spike_amplitude = None
        first_peak_idx = V[self.n_before : -self.n_after].argmax() + self.n_before
        first_spike = V[first_peak_idx - self.n_before : first_peak_idx + self.n_after]
        shape = spike_shape(
            first_spike,
            dt=1,
            deriv_thresh=deriv_thresh,
            t_baseline=self.n_before // 2,
            min_rise=self.n_rise // 4,
        )
        if shape is None or shape.takeoff_t is None:
            return
        spike_base = shape.takeoff_V
        first_spike_amplitude = shape.peak_V - spike_base
        self.spike_thresh = max(
            spike_base + thresh_rel * first_spike_amplitude,
            thresh_min,
        )
        return shape

    def detect_spikes(self, V: np.ndarray) -> Iterator[int]:
        """Using the calculated threshold, detect spikes that are extractable"""
        from quickspikes import detector, filter_times

        det = detector(self.spike_thresh, self.n_rise)
        return filter_times(det.send(V), self.n_before, V.size - self.n_after)

    def extract_spikes(
        self, V: np.ndarray, *, min_amplitude: float, upsample: int = 2, jitter: int = 4
    ) -> Iterator[Tuple[int, np.ndarray]]:
        """Detect and extract spikes from V.

        V: input signal
        min_amplitude: only yield spikes that are at least this much above threshold
        upsample: the upsampling factor to use in realigning spikes
        jitter: the expected jitter in peak position

        Yields (t, spike)
        """
        from quickspikes import peaks
        from quickspikes.tools import trim_waveforms, realign_spikes

        spike_times = self.detect_spikes(V)
        if len(spike_times) == 0:
            return
        spikes = peaks(V, spike_times, n_before=self.n_before, n_after=self.n_after)
        spike_times, spikes = realign_spikes(
            spike_times,
            spikes,
            upsample=upsample,
            expected_peak=self.n_before,
        )
        # original version of this located the peak after resampling, but I
        # don't think this is generally necessary
        for time, spike in trim_waveforms(
            spikes, spike_times, peak_t=self.n_before, n_rise=self.n_rise * 2
        ):
            if not np.any(spike - self.spike_thresh > min_amplitude):
                continue
            yield (time // upsample, spike)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""spike detection routines

Copyright (C) 2013 Dan Meliza <dmeliza@gmail.com>
Created Wed Jul 24 09:26:36 2013
"""
from .spikes import detector, peaks, subthreshold
from .tools import realign_spikes, filter_times

__version__ = "2.0.2"
__all__ = ["detector", "peaks", "subthreshold", "realign_spikes", "filter_times"]



# -*- coding: utf-8 -*-
# -*- mode: cython -*-
# cython: language_level=3
"""spike detection routines

Copyright (C) 2013 Dan Meliza <dmeliza@gmail.com>
Created Wed Jul 24 09:26:36 2013
"""
from cython cimport view, boundscheck
import numpy as np

ctypedef fused sample_t:
    short
    int
    long
    float
    double

cdef extern from "math.h":
    double NAN

cdef enum DetectorState:
    BELOW_THRESHOLD = 1
    BEFORE_PEAK = 2
    AFTER_PEAK = 3

cdef inline bint compare_sign(sample_t x, sample_t y):
    """return True iff ((x > 0) && (y > 0)) || ((x < 0) && (y < 0))"""
    return ((x >= 0) and (y >= 0)) or ((x < 0) and (y < 0))

cdef class detector:
    """Detect spikes in a continuous stream of samples.

    This implementation allows samples to be sent to the detector in blocks, and
    maintains state across successive calls.
    """
    cdef:
        double thresh
        double scaled_thresh
        double prev_val
        size_t n_after
        size_t n_after_crossing
        DetectorState state

    def __init__(self, double thresh, size_t n_after):
        """Construct spike detector.

        Parameters
        ----------
        thresh : double
          The crossing threshold that triggers the detector. NB: to detect
          negative-going peaks, invert the signal sent to send()
        n_after : size_t
          The maximum number of samples after threshold crossing to look for the
          peak. If a peak has not been located within this window, the crossing
          is considered an artifact and is not counted.

        """
        self.thresh = self.scaled_thresh = thresh
        self.n_after = n_after
        self.reset()

    def scale_thresh(self, double mean, double sd):
        """Adjust threshold for the mean and standard deviation of the signal.

        The effective threshold will be (thresh * sd + mean)

        """
        self.scaled_thresh = self.thresh * sd + mean

    def send(self, sample_t[:] samples):
        """Detect spikes in a time series.

        Returns a list of indices corresponding to the peaks in the data.
        Retains state between calls. Call reset() if there is a gap in the
        signal.

        """
        cdef double x
        cdef int i = 0
        out = []

        for i in range(samples.shape[0]):
            x = samples[i]
            if self.state is BELOW_THRESHOLD:
                if x >= self.scaled_thresh:
                    self.prev_val = x
                    self.n_after_crossing = 0
                    self.state = BEFORE_PEAK
            elif self.state is BEFORE_PEAK:
                if self.prev_val > x:
                    out.append(i - 1)
                    self.state = AFTER_PEAK
                elif self.n_after_crossing > self.n_after:
                    self.state = BELOW_THRESHOLD
                else:
                    self.prev_val = x
                    self.n_after_crossing += 1
            elif self.state is AFTER_PEAK:
                if x < self.scaled_thresh:
                    self.state = BELOW_THRESHOLD
        return out

    def __call__(self, double[:] samples):
        """Detect spikes in a time series.

        Returns a list of indices corresponding to the peaks in the data.
        Resets state between calls.

        """
        self.reset()
        return self.send(samples)

    def reset(self):
        """Reset the detector's internal state"""
        self.state = BELOW_THRESHOLD


@boundscheck(False)
def peaks(sample_t[:] samples, times, *, size_t n_before=75, size_t n_after=400):
    """Extracts samples around times

    Returns a 2D array with len(times) rows and (n_before + n_after) columns
    containing the values surrounding the sample indices in times.

    Note: all values of times must be greater than n_before and less than
    samples.size - n_after. See `tools.filter_spikes`

    """
    cdef size_t i = 0
    cdef size_t event, start, stop
    if sample_t is short:
        dtype = np.int16
    elif sample_t is int:
        dtype = np.int32
    elif sample_t is long:
        dtype = np.int64
    elif sample_t is float:
        dtype = np.float32
    elif sample_t is double:
        dtype = np.double
    out = np.empty(shape=(len(times), n_before + n_after), dtype=dtype)
    cdef sample_t [:, :] arr = out
    for i,event in enumerate(times):
        start = event - n_before
        stop = event + n_after
        if start < 0:
            raise ValueError("spike %d waveform starts before input data" % i)
        elif stop >= samples.size:
            raise ValueError("spike %d waveform ends after input data" % i)
        else:
            arr[i,:] = samples[event-n_before:event+n_after]

    return out


def find_run(sample_t[:] values, *, sample_t thresh, long min_run):
    """ Return the index of the first element in values that starts a run of at
    least min_run in length, or None if no such run exists. """
    cdef long i
    cdef long run_start = 0
    cdef bint in_run = False
    for i in range(values.size):
        if values[i] > thresh:
            if not in_run:
                in_run = True
                run_start = i
            if i - run_start >= min_run:
                return run_start
        else:
            in_run = False
    return None


def subthreshold(double[:] samples, times, *,
                 double v_thresh=-50, double dv_thresh=0, size_t min_size=10):
    """Removes spikes from time series

    Spikes are removed from the voltage trace by beginning at each peak and
    moving in either direction until V drops below thresh_v OR dv drops below
    thresh_dv. This algorithm does not work with negative peaks, so invert your
    signal if you need to.

    - samples : signal to analyze
    - times : times of the peaks (in samples)
    - thresh_v : do not include points contiguous with the peak with V > thresh_v
    - thresh_dv : do not include points contiguous with the peak with deltaV > thresh_dv.
                  negative values correspond to positive-going voltages
    - min_size : always remove at least this many points on either side of peak

    Returns a copy of samples with the data points around spike times set to NaN

    """
    cdef size_t i, j, spikestart, spikestop
    cdef size_t nsamples = samples.size
    cdef double[:] out = samples.copy()
    for i in times:
        if samples[i] < v_thresh:
            #print "spike peak at %d is below threshold; skipping"
            continue
        # iterate back and blank out samples before v or dv cross threshold
        j = 0
        spikestart = i
        while spikestart >= 0:
            if (j > min_size and
                (samples[spikestart + 1] - samples[spikestart] < dv_thresh) or
                (samples[spikestart] < v_thresh)):
                break
            out[spikestart] = NAN
            j += 1
            spikestart -= 1
        # iterate back and blank out samples before v or dv cross threshold
        j = 0
        spikestop = i
        while spikestop < nsamples:
            if (j > min_size and
                (samples[spikestop - 1] - samples[spikestop] < dv_thresh) or
                (samples[spikestop] < v_thresh)):
                break
            out[spikestop] = NAN
            j += 1
            spikestop += 1
    return out


# Variables:
# End:



# -*- coding: utf-8 -*-
# -*- mode: python -*-
import pytest

import numpy as np

from quickspikes.spikes import detector, peaks, find_run
from quickspikes.tools import filter_times, realign_spikes, trim_waveforms, peak_idx, trough_idx
from quickspikes.intracellular import SpikeFinder, spike_shape

# a nice surrogate spike with 20 samples before peak and 40 after
a_spike = np.array([-1290,  -483,  -136,  -148,  -186,   637,   328,    41,    63,
                    42,   377,   872,   639,   -17,   538,   631,   530,   693,
                    743,  3456,  6345,  5868,  4543,  3087,  1691,   830,   241,
                    -350,  -567,  -996,  -877, -1771, -1659, -1968, -2013, -2290,
                    -2143, -1715, -1526, -1108,  -500,   333,    25,  -388,  -368,
                    -435,  -817,  -858,  -793, -1089,   -16,  -430,  -529,  -252,
                    -3,  -786,   -47,  -266,  -963,  -365], dtype=np.int16)
t_peak = a_spike.argmax()
t_trough = a_spike.argmin()


@pytest.fixture
def extrac_times():
    return  [100, 400, 1200, 1500, 5000, 5200, 6123, 9730]


@pytest.fixture(params=[np.double, np.int32, np.int16])
def extrac_recording(extrac_times, request):
    a_recording = np.zeros(10000, dtype=request.param)
    for t in extrac_times:
        a_recording[t:t + a_spike.size] += a_spike
    return a_recording
    

@pytest.fixture
def extrac_spikes():
    np.random.seed(100)
    return np.tile(-1.0 * a_spike.astype("d"), (100, 1)) + np.random.randn(100, a_spike.size) * 100


@pytest.fixture(params=["wide", "narrow"])
def intrac_recording(request):
    # a bit klunky
    if request.param == "wide":
        recording = np.load("test/intra_spike.npy")
        times = [7635, 8412, 9363, 10424, 11447, 12661, 13887, 15079, 16373,
           17753, 19168, 20682, 22357, 23979, 25574, 27209, 28989,
           30508, 32088, 33778]
        takeoff = 24
    elif request.param == "narrow":
        recording = np.load("test/intra_spike_narrow.npy")
        times = [8325, 8816, 9368, 9985, 10619, 11276, 11968, 12610, 13240, 13900, 14485, 15193, 15840, 16601]
        takeoff = 14
    return recording, times, takeoff


def test_detect_extrac_spikes(extrac_recording, extrac_times):
    det = detector(2000, 40)
    assert det.send(extrac_recording) == [t + t_peak for t in extrac_times]
    assert det.send(-extrac_recording) == [t + t_trough for t in extrac_times]


def test_extract_spikes_nofilter(extrac_recording, extrac_times):
    with pytest.raises(ValueError):
        peaks(extrac_recording, [t + t_peak for t in extrac_times], n_before=20, n_after=300)


def test_extract_spikes(extrac_recording, extrac_times):
    n_before = 20
    n_after = 300
    times = filter_times([t + t_peak for t in extrac_times], n_before, extrac_recording.size - n_after)
    x = peaks(extrac_recording, times, n_before=n_before, n_after=n_after)
    # last peak is too close to end and should get dropped
    assert x.shape[0] == len(extrac_times) - 1
    assert x.shape[1] == n_before + n_after
    assert np.all(a_spike == x[0,:a_spike.size])


def test_align_extrac_by_peak(extrac_spikes):
    upsample = 3
    jitter = 3
    nevents, npoints = extrac_spikes.shape
    times, aligned = realign_spikes(np.zeros(nevents), extrac_spikes, upsample=upsample, jitter=jitter, align_by=peak_idx)
    apeak = aligned.argmax(-1)
    assert all(apeak == apeak[0])
    assert all(abs(times) <= jitter * upsample)


def test_align_extrac_by_trough(extrac_spikes):
    upsample = 3
    jitter = 3
    nevents, npoints = extrac_spikes.shape
    times, aligned = realign_spikes(np.zeros(nevents), extrac_spikes, upsample=upsample, jitter=jitter, align_by=trough_idx)
    apeak = aligned.argmin(-1)
    assert all(apeak == apeak[0])
    assert all(abs(times) <= jitter * upsample)
            

def test_detect_intrac_spikes(intrac_recording):
    recording, times, *_ = intrac_recording
    det = detector(-20, 100)
    assert det.send(recording) == times


def test_align_spikes(intrac_recording):
    recording, times, *_ = intrac_recording
    jitter = 4
    spikes = peaks(recording, times, n_before=200, n_after=400)
    aln_times, aligned = realign_spikes(times, spikes, upsample=3, jitter=jitter)
    apeak = aligned.argmax(1)
    assert all(apeak == apeak[0])
    # times should have shifted no more than jitter
    assert len(times) == aln_times.size
    for t1, t2 in zip(times, aln_times):
        assert abs(t1 - t2 // 3) <= jitter


def test_trim(intrac_recording):
    """spikes can be trimmed to eliminate overlapping peaks """
    recording, times, *_ = intrac_recording
    det = detector(-20, 100)
    spikes = peaks(recording, times, n_before=200, n_after=700)
    for _, spike_w in trim_waveforms(spikes, times, peak_t=200, n_rise=100):
        t = det(spike_w.astype("d"))
        assert t == [200]


def test_extrac_shape():
    shape = spike_shape(a_spike, 1)
    assert shape.peak_t == t_peak
    assert t_peak + shape.trough_t == t_trough


def test_intrac_trough(intrac_recording):
    """spike_shape finds trough under normal conditions"""
    recording, times, *_ = intrac_recording
    spikes = peaks(recording, times, n_before=200, n_after=400)
    for i, spike in enumerate(spikes):
        shape = spike_shape(spike, dt=1, t_baseline=100, min_rise=13)
        peak = shape.peak_t
        assert shape.trough_t == spike[peak:].argmin()


def test_intrac_trough_no_min(intrac_recording):
    """spike_shape finds trough when the spike is clipped"""
    recording, times, *_ = intrac_recording
    spikes = peaks(recording, times[:1], n_before=100, n_after=100)
    for spike in spikes:
        shape = spike_shape(spike, dt=1, t_baseline=100, min_rise=13)
        peak = shape.peak_t
        assert shape.trough_t == spike[peak:].argmin()


def test_intrac_onset(intrac_recording):
    """correctly detect onset time """
    # this case is based on manual inspection of the spike waveform
    recording, times, takeoff = intrac_recording
    spike = peaks(recording, times[:1], n_before=200, n_after=100)[0]
    shape = spike_shape(spike, dt=1, t_baseline=100, min_rise=13)
    assert shape.takeoff_t == takeoff


def test_intrac_no_onset(intrac_recording):
    recording, times, takeoff = intrac_recording
    spike = peaks(recording, times[:1], n_before=20, n_after=100)[0]
    shape = spike_shape(spike, dt=1, t_baseline=80, min_rise=13)
    assert shape.takeoff_t is None


def test_bad_spike_peak_at_edge():
    """ spike_shape returns None if max is at the edge of the waveform """
    bad = np.arange(100, -100, -1)
    assert spike_shape(bad, dt=1, t_baseline=100, min_rise=13) is None
    assert spike_shape(bad[::-1], dt=1, t_baseline=100, min_rise=13) is None


def test_spike_finder_intrac(intrac_recording):
    recording, times, takeoff = intrac_recording
    detector = SpikeFinder(50, 350, 5000)
    shape = detector.calculate_threshold(recording)
    assert shape.takeoff_t == takeoff
    for i, (time, spike) in enumerate(
            detector.extract_spikes(recording, min_amplitude=10, upsample=2, jitter=4)
    ):
        assert time == pytest.approx(times[i], abs=4)


def test_no_spikes():
    """ extract_spikes should yield nothing if there are no spikes detected """
    # In some rare cases, there will be a single spike that passes the first spike
    # threshold but is not detected due to the minimum height requirement
    # (thresh_min).
    detector = SpikeFinder(50, 350, 5000)
    signal = np.random.randn(100000)
    detector.spike_thresh = 50
    times = [t for t, s in detector.extract_spikes(signal, min_amplitude=40)]
    assert len(times) == 0


def test_thresh_for_max_at_edge_of_signal(intrac_recording):
    """ calculate_threshold should ignore spikes that are too close to the edge of the signal """
    recording, times, takeoff = intrac_recording
    detector = SpikeFinder(50, 350, 5000)
    clip = times[0] - 300
    _ = detector.calculate_threshold(recording[clip:])
    peaks = detector.detect_spikes(recording[clip:])
    assert peaks[0] + clip == times[1]


def test_run():
    a = np.arange(-10, 10)
    assert find_run(a, thresh=0, min_run=5) == 11


def test_no_run():
    a = np.zeros(20)
    a[10:14] = 1
    assert find_run(a, thresh=0, min_run=5) is None





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- mode: python -*-
from setuptools import setup, Extension
import numpy

setup(
    ext_modules=[Extension("quickspikes.spikes", sources=["quickspikes/spikes.pyx"])],
    include_dirs=[numpy.get_include()],
)
